/* ------------------------------------------------------------------------	*/
/*	INIFILE.C	-- .INI file handling functions								*/
/*																			*/
/*	The format of an .INI file is expected to be:							*/
/*	[progname]																*/
/*	VAR1=ANYSTRING															*/
/*	VAR2=ANOTHERARBITRARYSTRING												*/
/*	...																		*/
/*	[otherprg]																*/
/*	VAR3=ETC																*/
/*	...																		*/
/*																			*/
/*	The limit on string lengths including name and = is 1024 characters.	*/
/*																			*/
/*	If you work on a compiler or OS where sizeof(char) != 1, you will need	*/
/*	to modify the malloc()'s.												*/
/* ------------------------------------------------------------------------	*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>							/* For _argv				*/
#include <io.h>
char *fqfn( char *inname, char *outname );  /* Can't #include fioplus.h */

struct string                               /* String in a linked list  */
{
	struct string *next;					/* Next string				*/
	char text[];							/* Dyn allocated text area	*/
};

typedef struct
{
	char *fname;							/* FQFN of .INI file		*/
	char *pname;							/* Section label			*/
	struct string *strs;					/* First string				*/
	char dirty;								/* True if modified			*/
}
INI;										/* .INI file structure		*/

#define INIEXT		".INI"
#define MAXLEN		1024

/* ------------------------------------------------------------------------	*/
/*	INI *openini(const char *ininame, const char *progname)					*/
/*																			*/
/*	Opens an .INI file.  "ininame" is the .INI file name, the path and the	*/
/*	extension .INI are optional.  If no path is given, the following search	*/
/*	order is used:															*/
/*																			*/
/*	1. The path given in the INIFILES environment variable					*/
/*	2. The directory the program was started from (DOS 3.0 or newer)		*/
/*	3. The current directory												*/
/*																			*/
/*	The contents of the [progname] section is copied into memory generated	*/
/*	by malloc().  If there is any error, openini() returns NULL.			*/
/*																			*/
/*	The file is created if it does not exist, and a (blank) [progname]		*/
/*	section is appended if it did not already exist in the file.			*/
/* ------------------------------------------------------------------------ */

INI *openini(const char *ininame, const char *progname)
{
	FILE *fptr;
	INI *ini;
	char file[MAXLEN+1], fqfile[80], *cptr;
	struct string *str, **sptr;
	int found, inchr;


	if ( (ini = malloc(sizeof(INI))) == NULL )
		return NULL;

	file[0] = '\0';

	if ( strchr(ininame,'\\') == NULL || strchr(ininame,':') == NULL )
	{
		if ( (cptr = getenv("INIFILES")) != NULL )
		{
			strcpy(file,cptr);
			cptr = strchr(file,0);

			if ( cptr[-1] != '\\' && cptr[-1] != ':' )
			{
				cptr[0] = '\\';        
				cptr[1] = '\0';
			}
		}
		else if ( _osmajor >= 3 )
		{
			strcpy(file,_argv[0]);
			cptr = strrchr(file,'\\');
			cptr[1] = '\0';
		}
	}

	strcat(file,ininame);

	if ( strchr(file,'.') == NULL )
		strcat(file,INIEXT);

	fqfn(file,fqfile);

	if ( ( ini->fname = malloc(strlen(fqfile)+strlen(progname)+2) ) == NULL )
		return NULL;

	strcpy(ini->fname,fqfile);
	ini->pname = strchr(ini->fname,'\0')+1;
	strcpy(ini->pname,progname);
	ini->strs = NULL;
	ini->dirty = 0;
	sptr = &(ini->strs);

	if ( ( fptr = fopen(fqfile,"a+t") ) == NULL )
		return NULL;

	rewind(fptr);

	found = 0;

	while ( found == 0 )
	{
		inchr = getc(fptr);

		switch(inchr)
		{
			case '[':
						fgets(file,MAXLEN+1,fptr);
						if ( ( cptr = strchr(file,']') ) != NULL )
						{
							*cptr = '\0';

							if ( stricmp(file,progname) == 0 )
								found = 1;
						}
						break;

			case EOF:
						found = EOF;
						break;

			default:
						while ( inchr != '\n' && inchr != EOF )
						{
							inchr = getc(fptr);
						}
						if ( inchr == EOF )
						{
							putc('\n',fptr);
							found = EOF;
						}
						break;
		}
	}

	if ( found == EOF )
	{
		fprintf(fptr,"[%s]\n",progname);
		fclose(fptr);
		return ini;
	}

	while( 1 )							/* Loop until break				*/	
	{	
		if ( fgets(file,MAXLEN+1,fptr) == NULL )
		{
			break;
		}

		if ( ( cptr = strchr(file,'\n') ) != NULL )
			*cptr = '\0';

		if ( file[0] == '[' )
		{
			break;
		}

		found = strlen(file)+1+sizeof(struct string);
		
		if ( ( str = malloc(found) ) == NULL )
		{
			break;
		}

		*sptr = str;
		str->next = NULL;
		strcpy(str->text,file);
		sptr = &(str->next);
	}

	fclose(fptr);
	return ini;
}


/* ------------------------------------------------------------------------	*/
/*	void flushini(INI *ini)													*/
/*																			*/
/*	Updates an .INI file without closing it (although the physical file is	*/
/*	closed).  The fille is not updated if unchanged.						*/
/* ------------------------------------------------------------------------ */

void flushini(INI *ini)
{
	FILE *in, *out;
	char tmpname[80], *ext, line[MAXLEN+1];
	struct string *str;
	int i, found, ignore;

	if ( ini->dirty )
	{
		strcpy(tmpname,ini->fname);

		ext = strchr(tmpname,'.');
		if ( ext == NULL )
		{
			ext = strchr(tmpname,'\0');
			strcpy(ext,INIEXT);
		}
		ext++;

		i = 0;
		do
		{
			itoa(i++,ext,32);
		}
		while( access(tmpname,0) != -1 );
			
		out = fopen(tmpname,"wt");
		in = fopen(ini->fname,"rt");


		found = ignore = 0;

		while( fgets(line,MAXLEN+1,in) != NULL )
		{
			if ( line[0] == '[' )
			{
				ext = strchr(line,']');
				if ( ext != NULL )
				{
					*ext = '\0';
					
					if ( stricmp(line+1,ini->pname) == 0 )
					{
						if ( !found )
						{							
							*ext = ']';
							fputs(line,out);
			
							str = ini->strs;

							while ( str != NULL )
							{
								fputs(str->text,out);
								fputc('\n',out);
								str = str->next;
							}
						}
						found = ignore = 1;
					}
					else
					{
						*ext = ']';
						ignore = 0;
					}
				}
			}

			if ( !ignore )
				fputs(line,out);
		}

		fclose(in);
		fclose(out);

		remove(ini->fname);
		rename(tmpname,ini->fname);

		ini->dirty = 0;
	}
}



/* ------------------------------------------------------------------------	*/
/*	void closeini(INI *ini)													*/
/*																			*/
/*	Closes an .INI file, rewriting modified data if needed.					*/
/* ------------------------------------------------------------------------	*/

void closeini(INI *ini)
{
	struct string *str, *str1;

	flushini(ini);
	
	str = ini->strs;
	while ( str != NULL )
	{
		str1 = str->next;
		free(str);
		str = str1;
	}

	free(ini->fname);	
	free(ini);
}


/* ------------------------------------------------------------------------	*/
/*	char *getini(const char *vname, INI *ini)								*/
/*																			*/
/*	Returns a pointer to the given variable as a null-terminated string.	*/
/*	NULL is returned if the variable is undefined.							*/
/*																			*/
/*	PEStr and EStr are file-level statics since they are used by putini().	*/
/* ------------------------------------------------------------------------	*/

static struct string *EStr, **PEStr;

char *getini(const char *vname, INI *ini)
{
	char *equ;

	PEStr = &(ini->strs);

	while ( (EStr = *PEStr) != NULL )
	{
		equ = strchr(EStr->text,'=');

		if ( equ != NULL )
		{
			*equ = '\0';

			if ( stricmp(vname,EStr->text) == 0 )
			{
				*equ = '=';
				return equ+1;
			}
			*equ = '=';
		}

		PEStr = &(EStr->next);
	}

	return NULL;
}



/* ------------------------------------------------------------------------ */
/*	int putini(const char *setstr, INI *ini)								*/
/*																			*/
/*	Adds a new string to the INI string buffer.  The syntax of setstr is:	*/
/*																			*/
/*	"VAR=DATA"	to set VAR to DATA											*/
/*	"VAR="		to set VAR to an explicit empty value						*/
/*	"VAR"		to remove VAR from the .INI file completely					*/
/*																			*/
/*	Note that the syntax differs from putenv().  Also, the string is		*/
/*	copied in full, so a local value may be passed as setstr.  The putstr()	*/
/*	command may cause the INI string buffer to expand.						*/
/* ------------------------------------------------------------------------	*/

int putini(const char *setstr, INI *ini)
{
	struct string *str;
	char *equ;

	if ( ( equ = strchr(setstr,'=') ) != NULL )
	{
		if ( ( str = malloc(strlen(setstr)+1+sizeof(struct string)) ) == NULL )
		{
			return -1;
		}

		*equ = '\0';
	}

	getini(setstr,ini);

	if ( EStr != NULL )
	{
		*PEStr = EStr->next;
		free(EStr);
	}

	if ( equ != NULL )
	{
		*equ = '=';

		str->next = ini->strs;
		ini->strs = str;

		strcpy(str->text,setstr);
	}

	ini->dirty = 1;

	return 0;
}
